
#if __LINUX_X86__
	#include <stdio.h>
	#include <unistd.h>
	#include <sys/uio.h>
	#include <sys/ioctl.h>
	#include <sys/time.h>
	#include <errno.h>
	#include <pthread.h>
	#include <fcntl.h>
#endif

#include "XC_Engine.h"
#include "UnLinker.h"


class FSurfaceInfo;
class FTransTexture;
class FSurfaceFacet;
#include "UnRenDev.h" //Avoid including UnRender.h to reduce compile times
#include "XC_Networking.h"
#include "UnCon.h"
#include "UnXC_Lev.h"
#include "UnXC_Col.h"
#include "UnXC_Travel.h"
#include "XC_LZMA.h"

IMPLEMENT_CLASS(UXC_Level);

//Because I don't want to add extra includes
//Need to get rid of this

struct FBatch
{
	FBatch* NextBatch;
	INT Count;
	INT Size;
	INT Indices[4096]; //Size is variable

	FBatch( FBatch* InNext=NULL)
		: NextBatch(InNext)
		, Count(0)
	{}
	
	static FBatch* AllocateBatch( INT BatchSize, FBatch* Other=NULL)
	{
		guard(FBatch::AllocateBatch);

		BatchSize = Clamp( BatchSize, 32, 4096);
		FBatch* Result = (FBatch*)appMalloc( 12 + BatchSize * 4, TEXT("FBatch"));
		Result->NextBatch = Other;
		Result->Count = 0;
		Result->Size = BatchSize;
		return Result;

		unguard;
	}
};

struct FNameBatchMap : public TMap<INT,FBatch*>
{
	TArray<INT> TransientNames;
	
	INT GetNameHash( const TCHAR* InStr)
	{
		guard(FNameBatchMap::GetNameHash);
		INT Len = Min( appStrlen(InStr), 63);
		TCHAR NewName[64];
		NewName[Len--] = 0;
		while ( (Len > 0) && appIsDigit(InStr[Len]) )
			NewName[Len--] = 0;
		for ( ; Len >= 0 ; Len-- )
			NewName[Len] = InStr[Len];
		return appStrihash( NewName) & 0x7FFFFFFF;
		unguard;
	}
	FName Request( UObject* Parent, UClass* ForClass )
	{
		guard(FNameBatchMap::Request);
		INT iHash = GetNameHash( ForClass->GetName() );
		if ( !Hash )
			Rehash();
		FBatch* Batch = FindRef( iHash);
		while ( Batch )
		{
			while ( Batch->Count > 0 )
			{
				//If object matches class and is deleted, pick it up
				Batch->Count--;
				FNameEntry* Entry = FName::GetEntry( Batch->Indices[Batch->Count] );
				if ( Entry && (Entry->Flags & RF_Transient) && (GetNameHash( Entry->Name) == iHash) && !UObject::StaticFindObject( NULL, Parent, Entry->Name ) )
					return *((FName*)&Entry->Index);
			}
			FBatch* NextBatch = Batch->NextBatch;
			delete Batch;
			Batch = NextBatch;
			if ( Batch )
			{
//				debugf( TEXT("Next batch in queue: (%i/%i)"), Batch->Count, Batch->Size);
				Set( iHash, Batch);
				check( FindRef(iHash) );
			}
			else
				Remove( iHash);
		}
		return NAME_None;
		unguard;
	}

	void Store( FName AName)
	{
		guard(FNameBatchMap::Store);
		INT iHash = GetNameHash( *AName);
		FBatch* Batch = FindRef( iHash );
		guard(BatchSetup);
		if ( !Batch )
		{
			Batch = FBatch::AllocateBatch(32);
			Set( iHash, Batch);
		}
		else if ( Batch->Count >= Batch->Size )
		{
			Batch = FBatch::AllocateBatch( Batch->Size + (Batch->Size >> 2) + (iHash%8), Batch);
			Set( iHash, Batch);
		}
		check( Batch->Count < Batch->Size );	
		unguard;
		//Add single name
		Batch->Indices[Batch->Count++] = *(INT*)&AName;
		unguard;
	}
	
	void Cleanup()
	{
		guard(FNameBatchMap::Cleanup);
		INT i;
		for ( i=0 ; i<TransientNames.Num() ; i++ )
		{
			FNameEntry* NEntry = FName::GetEntry(TransientNames(i));
			if ( NEntry )
				NEntry->Flags &= ~RF_Destroyed;
		}
		SafeEmpty(TransientNames);
/*		INT MaxNames = FName::GetMaxNames();
		for ( i=0 ; i<MaxNames ; i++ )
		{
			FNameEntry* NEntry = FName::GetEntry(i);
			if ( NEntry )
				NEntry->Flags &= ~RF_Destroyed;
		}
*/		for ( i=0 ; i<Pairs.Num() ; i++ )
		{
			FBatch* Batch = Pairs(i).Value;
			while ( Batch )
			{
				FBatch* Next = Batch->NextBatch;
				delete Batch;
				Batch = Next;
			}
		}
		Empty();
		unguard;
	}
};

static INT BatchData[8] = { 0, 0, 0, 0, 8, 0, 0, 0 };
static FNameBatchMap* UnusedNamePool = (FNameBatchMap*)BatchData;

//Stats
static INT ST_Traces = 0;

//Map loader stuff
static volatile INT GMapLoadCount = 0;
static volatile UBOOL GIsLoadingMap = 0; //Needed to avoid critical errors
static volatile UBOOL GLockPackageMap = 0;

//Throw warning to compiler if one of these hacks is missing
inline void CompilerCheck()
{
/** Add these extra bitfield values in AActor as follows:

    BITFIELD bClientDemoNetFunc:1;
	//
	BITFIELD bNotRelevantToOwner:1; //Lag compensators love this
	BITFIELD bRelevantIfOwnerIs:1; //Good for player attachments
	BITFIELD bRelevantToTeam:1;
	BITFIELD bTearOff:1;
	BITFIELD bNetDirty:1;
	//
    class UClass* RenderIteratorClass GCC_PACK(4);

*/
	AActor* TestActor;
	TestActor->bNotRelevantToOwner = TestActor->bNotRelevantToOwner;
	TestActor->bRelevantIfOwnerIs = TestActor->bRelevantIfOwnerIs;
	TestActor->bRelevantToTeam = TestActor->bRelevantToTeam;
	TestActor->bSuperClassRelevancy = TestActor->bSuperClassRelevancy;
	TestActor->bTearOff = TestActor->bTearOff;
	TestActor->bNetDirty = TestActor->bNetDirty;
}


enum EOwned { EPri_Owned };
enum EGeneral { EPri_General };


struct FActorPriority
{
	INT			Priority;	// Update priority, higher = more important.
	AActor*			Actor;		// Actor.
	UActorChannel*	Channel;	// Actor channel.
	UBOOL		SuperRelevant; //Skip owner checks later
	
	FActorPriority()
	{}
	//Use this on owned actors (NetOwner)
	FActorPriority( EOwned, FVector4& Location, FVector4& Dir, UNetConnection* InConnection, UActorChannel* InChannel, AActor* Target)
	{
		Actor = Target;
		Channel = InChannel;
		if ( Actor->bNotRelevantToOwner )
		{
			SuperRelevant = -1;
			Priority = -3000000;
			return;
		}
		FLOAT Time  = Channel ? (InConnection->Driver->Time - Channel->LastUpdateTime) : InConnection->Driver->SpawnPrioritySeconds;
		FLOAT Dot = (Actor == InConnection->Actor) ? 4.f : 1.5f;
		SuperRelevant = 1;
		if ( !Actor->bAlwaysRelevant )
		{
			if ( InConnection->Actor->Weapon == Actor )
				Dot += 1.0f;
			else if ( (Actor->RemoteRole > ROLE_DumbProxy) || (Actor->Physics != PHYS_None) )
			{
				FVector4 Delta( Actor->Location - Location);
				Dot += (Delta.Dot4(Dir) >= 0) ? 1.f : -1.f;  //Linux version doesn't have ASM, safe to use FVector4 local
			}
		}

		AActor* Sent = (Channel && Channel->Recent.Num()) ? (AActor*) &Channel->Recent(0) : NULL;
		FLOAT Pri = Actor->GetNetPriority( Sent, Time, InConnection->BestLag);
		Priority = appRound( (Dot + 3.0f) * Pri * 65536.f);

		if ( Actor->bNetOptional )
			Priority -= 3000000;
	}
	//Don't use this on owned actors (NetOwner)
	FActorPriority( EGeneral, FVector4& Location, FVector4& Dir, UNetConnection* InConnection, UActorChannel* InChannel, AActor* Target)
	{
		Actor = Target;
		Channel = InChannel;
		FLOAT Time  = Channel ? (InConnection->Driver->Time - Channel->LastUpdateTime) : InConnection->Driver->SpawnPrioritySeconds;
		FLOAT Dot = 0.f;
		SuperRelevant = 0;

		if ( !Actor->bAlwaysRelevant )
		{
			if ( (Actor->RemoteRole > ROLE_DumbProxy) || (Actor->Physics != PHYS_None) )
			{
				FVector4 Delta( Actor->Location - Location);
				Dot = (Delta.Dot4(Dir) >= 0) ? 1.f : -1.f;  //Linux version doesn't have ASM, safe to use FVector4 local
			}
		}
		APlayerPawn* NetOwner = InConnection->Actor;
		AActor* Sent = (Channel && Channel->Recent.Num()) ? (AActor*) &Channel->Recent(0) : NULL;

		if ( Actor == NetOwner->ViewTarget )
		{
			SuperRelevant = 1;
			Dot = 2.0f;
		}
		else
			SuperRelevant = Actor->bAlwaysRelevant || (NetOwner->ViewTarget && (Actor->IsOwnedBy(NetOwner->ViewTarget) || Actor->Instigator == NetOwner->ViewTarget));

		FLOAT Pri = Actor->GetNetPriority( Sent, Time, InConnection->BestLag);
		Priority = appRound( (Dot + 3.0f) * Pri * 65536.f);

		if ( Actor->bNetOptional )
			Priority -= 3000000;
	}
	friend INT Compare( const FActorPriority* A, const FActorPriority* B )
	{
		return B->Priority - A->Priority;
	}
};

//Fix this if old relevancy was toggled
//If True, actor is to be globally checked
inline FLOAT UpdateNetTag( AActor* Other, FLOAT DeltaTime)
{
	Other->NetTag &= 0x7FFFFFFF; //Become positive
	FLOAT* NetTag = (FLOAT*) &Other->NetTag;
	if ( *NetTag > 120.f || appIsNan(*NetTag) )
		*NetTag = appFrand() * 99.f;
	else if ( *NetTag > 99.f )
		*NetTag -= 99.f;
	FLOAT TestVal = *NetTag;
	*NetTag += DeltaTime;
	return TestVal;
}

inline UBOOL OwnedByAPlayer( const AActor* Test)
{
	for ( ; Test ; Test=Test->Owner )
		if ( Test->IsA(APlayerPawn::StaticClass()) && Cast<UNetConnection>(((APlayerPawn*)Test)->Player) )
			return true;
	return false;
}

//See if actor is relevant to a player
static UBOOL ActorIsRelevant( AActor* Actor, APlayerPawn* Player, AActor* Viewer, FVector4* ViewPos, const FVector4& EndOffset)
{
//	if ( Actor->bAlwaysRelevant ) //Handled by super relevant
//		return 1;
	if ( Actor->AmbientSound && ((*ViewPos-Viewer->Location).SizeSquared() < Square(25.0f*Actor->SoundRadius)) )
		return 1;
	if ( Actor->Owner )
	{
		if ( Actor->IsOwnedBy(Viewer) )
			return 1;
		if ( Actor->Owner->bIsPawn && ((APawn*)Actor->Owner)->Weapon == Actor )
			return ActorIsRelevant( Actor->Owner, Player, Viewer, ViewPos, EndOffset);
	}
	if ( (Actor->bHidden || Actor->bOnlyOwnerSee) && !Actor->bBlockPlayers && !Actor->AmbientSound )
		return 0;
	ST_Traces++;
#if ASM
	FVector4 Dest;
	__asm
	{
		mov      eax,[Actor]
		mov      esi,[EndOffset]
		lea      ecx,[Dest]
		movaps   xmm0,[esi]			//x0: EndOffset
		add      eax,208			//Move towards Actor->Location
		movups   xmm2,[eax]			//x2: Location
		add      eax,192			//Move towards Actor->CollisionRadius
		movlps   xmm3,[eax]			//x3: |CR|CH|..|..|
		shufps   xmm3,xmm3,16		//x3: |CR|CR|CH|CR| >>> 0b00010000
		mulps    xmm0,xmm3			//x0: Cylinder offset
		addps    xmm0,xmm2			//x0: Location + Cylinder offset
		movaps   [ecx],xmm0			//Store in Dest
	}
	return ((UXC_Model*) Viewer->XLevel->Model)->FastRepLineCheck( ViewPos, &Dest);
#elif ASMLINUX
	GCC_FVector4_ref(Dest); //GCC isn't aligning FVector4 locals
	__asm__ __volatile__(
	"movups    (%%esi),%%xmm0 \n" //x0: EndOffset |||| THIS IS LIKELY UNALIGNED!!!
	"addl         $208,%%eax  \n" //Move towards Actor->Location
	"movups    (%%eax),%%xmm2 \n" //x2: Location
	"addl         $192,%%eax  \n" //Move towards Actor->CollisionRadius
	"movlps    (%%eax),%%xmm3 \n" //x3: |CR|CH|..|..|
	"shufps $16,%%xmm3,%%xmm3 \n" //x3: |CR|CR|CH|CR| >>> 0b00010000
	"mulps      %%xmm3,%%xmm0 \n" //x0: Cylinder offset
	"addps      %%xmm2,%%xmm0 \n" //x0: Location + Cylinder offset
	"movaps     %%xmm0,(%%ecx) \n" //Store in Dest
	: : "a"(Actor), "c"(Dest), "S"(&EndOffset) : "memory" );
	return ((UXC_Model*) Viewer->XLevel->Model)->FastRepLineCheck( ViewPos, Dest);
#else
	FVector4 Dest;
	Dest.X = Actor->Location.X + EndOffset.X * Actor->CollisionRadius;
	Dest.Y = Actor->Location.Y + EndOffset.Y * Actor->CollisionRadius;
	Dest.Z = Actor->Location.Z + EndOffset.Z * Actor->CollisionHeight;
	return ((UXC_Model*) Viewer->XLevel->Model)->FastRepLineCheck( ViewPos, &Dest);
#endif
}

//BROWSE IS CALLED TOWARDS LOCALMAPURL DURING INIT, WHICH THEN CALLS LOADMAP!
UBOOL UXC_GameEngine::Browse( FURL URL, const TMap<FString,FString>* TravelInfo, FString& Error )
{
	//We're initializing UGameEngine, we no longer need this hack
	if ( UXC_GameEngine::StaticClass()->PropertiesSize < UGameEngine::StaticClass()->PropertiesSize )
		Exchange( UGameEngine::StaticClass()->PropertiesSize, UXC_GameEngine::StaticClass()->PropertiesSize );

	UBOOL bDisconnected = URL.HasOption( TEXT("failed")) || URL.HasOption( TEXT("entry"));
	if ( bDisconnected )
	{
		ULevel* MyLevel = (&GLevel)[b451Setup];
		if ( MyLevel )
			MyLevel->SetActorCollision(0);
		UnHookEngine();
	}

	if ( bEnableDebugLogs )
		debugf( NAME_XC_Engine, TEXT("Browse() Start: %s %s %i"), *URL.Protocol, *URL.Host, URL.Port);
	UBOOL result = Super::Browse( URL, TravelInfo, Error);
	if ( bEnableDebugLogs )
		debugf( NAME_XC_Engine, TEXT("Browse() End"));

	if ( bDisconnected || !GetActiveLevel(this) )
	{
		HookNatives( true); //Full GNative hooks
		HookEngine( NULL ); //Forces limited hook of standard stuff
		AdminLoginHook = NULL;
//		CollectGarbage(RF_Native); //EXPERIMENTAL
	}

	return result;
}


/////////////////////
// Compressor thread
/////////////////////
struct FAutoCompressorThread
{
public:
	// Variables.
	DWORD			tId;
	UPackageMap*	PackageMap;
	INT				MapCount;
	TCHAR			CompressedExt[12];
#if __UNIX__
	pthread_t		ACThread;
#endif

#if _MSC_VER
	HANDLE			hThread;
#endif

	// Functions.
	FAutoCompressorThread()	{}
	FAutoCompressorThread( UPackageMap* InPackageMap, INT InMapCount=GMapLoadCount);
};



// Priority thread entrypoint.
#if __UNIX__
void* AutoCompressorThreadEntry(void* Arg)
#else
DWORD STDCALL AutoCompressorThreadEntry(void* Arg)
#endif
{
	#ifdef _MSC_VER
	appSleep( 0.2f);
	FAutoCompressorThread* ACT = (FAutoCompressorThread*)Arg;
	INT LenExt = appStrlen( ACT->CompressedExt);

	INT i = 0;
	UBOOL bLeave = 0;
	while ( !bLeave )
	{
		appSleep( 0.002f);
		if ( GIsRequestingExit) //Game about to exit
			return 0;
		if ( FPlatformAtomics::InterlockedCompareExchange(&GMapLoadCount, 0, 0) != ACT->MapCount) //New map has loaded
			break;
		if ( FPlatformAtomics::InterlockedCompareExchange(&GIsLoadingMap, 0, 0) != 0) //Map is loading!!
		{
			appSleep( 0.1f);
			continue;
		}
		if ( FPlatformAtomics::InterlockedCompareExchange(&GLockPackageMap, 1, 0)) //Other thread is compressing
		{
			appSleep( 0.1f);
			continue;
		}
		if ( i < ACT->PackageMap->List.Num() )
		{
			ULinker* LNKh = (ULinker*)ACT->PackageMap->List(i).Linker;
			INT Len = LNKh->Filename.Len();
			if ( Len < 250 - LenExt ) //Avoid string formatting crash
			{
				TCHAR FileName_Base[256];
				appStrcpy( FileName_Base, *LNKh->Filename);
				TCHAR FileName_LZMA[256];
				appSprintf( FileName_LZMA, TEXT("%s%s"), FileName_Base, ACT->CompressedExt);


				if ( GFileManager->FileSize( FileName_LZMA) <= 0 )
				{	//Unsafe operations ended, let's unlock main thread's loaders
					FPlatformAtomics::InterlockedExchange(&GLockPackageMap, 0);
					TCHAR Error[256]; Error[0] = 0;
					TCHAR FileName_Temp[256];
					appStrcpy( FileName_Temp, FileName_LZMA);
					FileName_Temp[Len + LenExt - 1] = 't'; //TEMP INDICATOR
					TCHAR CMPMessage[600];
					appSprintf( CMPMessage, TEXT("AutoCompressing %s to %s"), FileName_Base, FileName_LZMA);
					ThreadedLog( NAME_Log, CMPMessage );
					LzmaCompress( FileName_Base, FileName_Temp, Error);
					GFileManager->Move( FileName_LZMA, FileName_Temp);
				}
			}
		}
		else
			bLeave = 1;
		i++;
		FPlatformAtomics::InterlockedExchange(&GLockPackageMap, 0);
	}
	#endif
	return 0;
}

FAutoCompressorThread::FAutoCompressorThread(UPackageMap* InPackageMap, INT InMapCount)
	:	PackageMap(InPackageMap)
	,	MapCount(InMapCount)
{
	guard(WTF);
#ifdef _MSC_VER
	appSprintf( CompressedExt, COMPRESSED_EXTENSION);
	FixFilename( CompressedExt); //Init safe File Manager if necessary

	debugf(NAME_Log, TEXT("Starting LZMA compressor...") );


	DWORD exitCode = 0;

	HANDLE hThread = CreateThread( NULL, 0, AutoCompressorThreadEntry, this, 0, &tId);
	debugf(NAME_XC_Engine, TEXT("Thread ID: %i"), tId);

	//Debugging code I want to keep
	check(hThread);
	GetExitCodeThread(hThread, &exitCode);
	ResumeThread(hThread);
	WaitForSingleObject(hThread, 1); //Waste 1 millisecond
	GetExitCodeThread(hThread, &exitCode);
	CloseHandle( hThread);
#endif
	unguard;
}

static UBOOL PackageNeedsDemoFix( const TCHAR* PkgName)
{
	guard( PackageNeedsDemoFix);
	TCHAR AnnouncerText[] = {'A','n','n','o','u','n','c','e','r', 0 };
	TCHAR CreditsText[] = {'C','r','e','d','i','t','s', 0 };
	const TCHAR* DemoPkgs[] = { AnnouncerText, CreditsText};

	for ( INT i=0 ; i<2 ; i++ )
		if ( appStricmp(DemoPkgs[i], PkgName) == 0 )
			return 1;
	return 0;
	unguard;
}

#include "XCL_Private.h"

void UXC_Level::CleanupDestroyed( UBOOL bForce )
{
	guard(UXC_Level::CleanupDestroyed);
	if ( bForce ) //Likely on level switch
	{
		Super::CleanupDestroyed( bForce);
		return;
	}

	// Pack actor list.
	CompactActors();

	// If nothing deleted, exit.
	if( !FirstDeleted )
		return;

	// Don't do anything unless a bunch of actors are in line to be destroyed.
	guard(CheckDeleted);
	INT c=0;
	for( AActor* A=FirstDeleted; A; A=A->Deleted )
		c++;
	if( c<128 )
		return;
	unguard;

	// Remove all references to actors tagged for deletion.
	guard(CleanupRefs);
	for( INT iActor=0; iActor<Actors.Num(); iActor++ )
	{
		AActor* Actor = Actors(iActor);
		if( Actor )
		{
			// Would be nice to say if(!Actor->bStatic), but we can't count on it.
			checkSlow(!Actor->bDeleteMe);
			Actor->GetClass()->CleanupDestroyed( (BYTE*)Actor );
		}
	}
	unguard;
	
	// Remove all references from grid if present
	guard(CleanupGrid);
	if ( CollisionCacus )
		CollisionCacus->Cleanup();
	unguard;
	
	guard(FinishDestroyedActors);
	while( FirstDeleted!=NULL )
	{
		// Physically destroy the actor-to-delete.
		check(FirstDeleted->bDeleteMe);
		AActor* ActorToKill = FirstDeleted;
		FirstDeleted        = FirstDeleted->Deleted;
		check(ActorToKill->bDeleteMe);

		FNameEntry* NEntry = FName::GetEntry( ActorToKill->GetFName().GetIndex() );
		if ( NEntry->Flags & RF_Transient ) //I tagged this name upon it's creation
			UnusedNamePool->Store( ActorToKill->GetFName() );
		
		// Destroy the actor.
		delete ActorToKill;
	}
	unguard;

	unguard;
}


//
// Create a new actor. Returns the new actor, or NULL if failure.
//
AActor* UXC_Level::SpawnActor
(
	UClass*			SpawnClass,
	FName			InName,
	AActor*			Owner,
	class APawn*	Instigator,
	FVector			Location,
	FRotator		Rotation,
	AActor*			Template,
	UBOOL			bNoCollisionFail,
	UBOOL			bRemoteOwned
)
{
	guard(UXC_Level::SpawnActor);

	// Make sure this class is spawnable.
	if( !SpawnClass)
	{
		debugf( NAME_Warning, TEXT("SpawnActor failed because no class was specified") );
		return NULL;
	}
	if(SpawnClass->ClassFlags & CLASS_Abstract )
	{
		debugf( NAME_Warning, TEXT("SpawnActor failed because class %s is abstract"), SpawnClass->GetName() );
		return NULL;
	}
	else if( !SpawnClass->IsChildOf(AActor::StaticClass()) )
	{
		debugf( NAME_Warning, TEXT("SpawnActor failed because %s is not an actor class"), SpawnClass->GetName() );
		return NULL;
	}
	else if( !GIsEditor && (SpawnClass->GetDefaultActor()->bStatic || SpawnClass->GetDefaultActor()->bNoDelete) )
	{
		debugf( NAME_Warning, TEXT("SpawnActor failed because class %s has bStatic or bNoDelete"), SpawnClass->GetName() );
		return NULL;		
	}

	// Use class's default actor as a template.
//	UBOOL bFromDefaults = (Template == NULL);
	if( !Template )
		Template = SpawnClass->GetDefaultActor();
	check(Template!=NULL);

	// Make sure actor will fit at desired location, and adjust location if necessary.
	if( (Template->bCollideWorld || (Template->bCollideWhenPlacing && (GetLevelInfo()->NetMode != NM_Client))) && !bNoCollisionFail )
		if( !FindSpot( Template->GetCylinderExtent(), Location, 0, 1 ) )
			return NULL;

	// Add at end of list.
	INT iActor = Actors.Add();
	AActor* Actor = NULL;
	
	guard(CreateActor);
	if ( InName == NAME_None )
		InName = UnusedNamePool->Request( GetOuter(), SpawnClass);
	Actor = Actors(iActor) = (AActor*)StaticConstructObject(SpawnClass, GetOuter(), InName, 0, Template );
	if ( InName == NAME_None ) //Name was created instead!!!
	{
		Actor->GetFName().SetFlags(RF_Transient);
		UnusedNamePool->TransientNames.AddItem( Actor->GetFName().GetIndex());
	}
	unguard;
	//VS 2015 has an unknown problem
/*	if (bFromDefaults && Actor->Role != ROLE_Authority)
	{
		BYTE ObjData[40];
		appMemcpy_amd(ObjData, Actor, 40);
		InitProperties((BYTE*)Actor, SpawnClass->GetPropertiesSize(), SpawnClass, NULL, SpawnClass->GetPropertiesSize());
		appMemcpy_amd(Actor, ObjData, 40);
	}*/
	Actor->SetFlags( RF_Transactional );

	// Set base actor properties.
	Actor->Tag		= SpawnClass->GetFName();
	Actor->Region	= FPointRegion( GetLevelInfo() );
	Actor->Level	= GetLevelInfo();
	Actor->bTicked  = !Ticked;
	Actor->XLevel	= this;

	check(Actor->Role==ROLE_Authority);
	if( bRemoteOwned )
		Exchange( Actor->Role, Actor->RemoteRole );

	// Remove the actor's brush, if it has one, because moving brushes are not duplicatable.
	if( Actor->Brush )
		Actor->Brush = NULL;

	// Set the actor's location and rotation.
	Actor->Location = Location;
	Actor->OldLocation = Location;
	Actor->Rotation = Rotation;
	if( Actor->bCollideActors && Hash  )
		Hash->AddActor( Actor );

	// Init the actor's zone.
	Actor->Region = FPointRegion(GetLevelInfo());
	if( Actor->IsA(APawn::StaticClass()) )
		((APawn*)Actor)->FootRegion = ((APawn*)Actor)->HeadRegion = FPointRegion(GetLevelInfo());

	// Set owner/instigator
	Actor->SetOwner( Owner );
	Actor->Instigator = Instigator;

	// Send messages.
	Actor->InitExecution();
	Actor->Spawned();
	Actor->eventSpawned();
	Actor->eventPreBeginPlay();
	Actor->eventBeginPlay();
	if( Actor->bDeleteMe )
		return NULL;

	// Set the actor's zone.
	SetActorZone( Actor, iActor==0, 1 );

	// Send PostBeginPlay.
	Actor->eventPostBeginPlay();

	// Check for encroachment.
	if( !bNoCollisionFail && CheckEncroachment( Actor, Actor->Location, Actor->Rotation, 0 ) )
	{
		DestroyActor( Actor );
		return NULL;
	}

	// Init scripting.
	Actor->eventSetInitialState();

	// Find Base
	if( !Actor->Base && Actor->bCollideWorld
		 && (Actor->IsA(ADecoration::StaticClass()) || Actor->IsA(AInventory::StaticClass()) || Actor->IsA(APawn::StaticClass())) 
		 && ((Actor->Physics == PHYS_None) || (Actor->Physics == PHYS_Rotating)) )
		Actor->FindBase();

	// Success: Return the actor.
	if( InTick )
		NewlySpawned = new(GEngineMem)FActorLink(Actor,NewlySpawned);

	static UBOOL InsideNotification = 0;
	if( !InsideNotification )
	{
		InsideNotification = 1;
		// Spawn notification
		for( ASpawnNotify* N = GetLevelInfo()->SpawnNotify; N; N = N->Next )
		{
			if( N->ActorClass && Actor->IsA(N->ActorClass) )
				Actor = N->eventSpawnNotification( Actor );
		}
		InsideNotification = 0;
	}

	return Actor;
	unguardf(( TEXT("(%s)"), SpawnClass->GetName() ));
}

void UXC_Level::SetActorCollision( UBOOL bCollision )
{
	guard(UXC_Level::SetActorCollision);

	UXC_GameEngine* XCGE = Cast<UXC_GameEngine>(Engine);
	if ( !XCGE || !XCGE->bCollisionHashHook )
	{
		Super::SetActorCollision(bCollision);
		return;
	}

	// Init collision if first time through.
	if( bCollision && !Hash )
	{
		// Init gridded octree.
		guard(StartCollision);
		debugf( NAME_XC_Engine, TEXT("Setting up collision octree [Level Hook] for %s"), GetOuter()->GetName() );
		CollisionCacus = new FCollisionCacus(this);
		Hash = CollisionCacus;
		for( INT i=0; i<Actors.Num(); i++ )
			if( Actors(i) && Actors(i)->bCollideActors )
				Hash->AddActor( Actors(i) );
		unguard;
	}
	else if( Hash && !bCollision )
	{
		// Destroy gridded octree, no need to remove actors.
		guard(EndCollision);
		if ( CollisionCacus == Hash )
		{
			CollisionCacus->Exit();
			delete CollisionCacus;
			Hash = NULL;
			CollisionCacus = NULL;
		}
		else
		{
			debugf( NAME_XC_Engine, TEXT("CollisionCacus not equal to %s hash!!!"), GetOuter()->GetName() );
			Super::SetActorCollision(bCollision);
		}
		unguard;
	}

	unguard;
}

void UXC_Level::Destroy()
{
	guard(ULevel::Destroy);

	// Free allocated stuff.
	SetActorCollision(0);

	if( BrushTracker )
	{
		delete BrushTracker;
		BrushTracker = NULL; /* Required because brushes may clean themselves up. */
	}

	ULevelBase::Destroy();
	unguard;
}

//
// Welcome a new player joining this server.
//
void UXC_Level::WelcomePlayer( UNetConnection* Connection, TCHAR* Optional )
{
	guard(ULevel::WelcomePlayer);

	Connection->PackageMap->Copy( Connection->Driver->MasterMap );
	Connection->SendPackageMap();
	Connection->Logf( TEXT("XC_ENGINE VERSION=%i"), ((UXC_GameEngine*)Engine)->Version ); //Find a way to make clients interpret this
	if( Optional[0] )
		Connection->Logf( TEXT("WELCOME LEVEL=%s LONE=%i %s"), GetOuter()->GetName(), GetLevelInfo()->bLonePlayer, Optional );
	else
		Connection->Logf( TEXT("WELCOME LEVEL=%s LONE=%i"), GetOuter()->GetName(), GetLevelInfo()->bLonePlayer );
	Connection->FlushNet();

	unguard;
}

void UXC_Level::TickNetServer( FLOAT DeltaSeconds )
{
	guard(UXC_Level::TickNetServer);

	if ( ! ((UXC_GameEngine*)Engine)->bUseNewRelevancy )
	{
		Super::TickNetServer( DeltaSeconds);
		return;
	}

	ThreadedLogFlush();
	// Update window title
	if( (INT)(TimeSeconds-DeltaSeconds)!=(INT)(TimeSeconds.GetFloat()) )
		debugf( NAME_Title, LocalizeProgress(TEXT("RunningNet"),TEXT("Engine")), *GetLevelInfo()->Title, *URL.Map, NetDriver->ClientConnections.Num() );
	
	clock(NetTickCycles);
	INT Updated = ServerTickClients( DeltaSeconds);
	unclock(NetTickCycles);

	// Stats.
	static INT SkipCount = 0;
	static FLOAT AccTickRate = 0;

	if ( Updated )
	{
		SkipCount++;
		AccTickRate += DeltaSeconds;
	}
	if( Updated && (20.f > appFrand()*Engine->CurrentTickRate) )
	{
		static INT NumActors;
		NumActors = 0;
		FLOAT TickRate = 0;
		for( INT i=0; i<NetDriver->ClientConnections.Num(); i++ )
		{
			UNetConnection* Connection = NetDriver->ClientConnections(i);
			if( Connection->Actor && Connection->State==USOCK_Open )
			{
				if( Connection->UserFlags&1 )
				{
					if ( !NumActors ) //Prevent multiple UserFlag clients from using extra CPU cycles
					{
						AccTickRate = GetLevelInfo()->TimeDilation * (FLOAT)SkipCount / AccTickRate;
						TickRate = Engine->GetMaxTickRate();
						NumActors = iFirstDynamicActor - 1; //No brush lol
						for( INT ii=iFirstDynamicActor; ii<Actors.Num(); ii++ )
							NumActors += Actors(ii)!=NULL;
					}

					TCHAR FPS_Str[16];
					appSprintf(FPS_Str, (TickRate < 100.f) ? TEXT("%04.1f") : TEXT("%05.1f"), AccTickRate);
					FString Stats = FString::Printf
					(
						TEXT("r=%s (%02.0f) cli=%i act=%03.1f (%i) net=%03.1f pv/c=%i rep/c=%i trace/c=%i"),
						FPS_Str,
						TickRate,
						NetDriver->ClientConnections.Num(),
						GSecondsPerCycle * 1000 * ActorTickCycles,
						NumActors,
						GXSecondsPerCycle * 1000 * NetTickCycles,
						NumPV/NetDriver->ClientConnections.Num(),
						NumReps/NetDriver->ClientConnections.Num(),
						ST_Traces/(NetDriver->ClientConnections.Num() * SkipCount)
					);
					Connection->Actor->eventClientMessage( *Stats, NAME_None, 0 );

				}
				if( Connection->UserFlags&2 )
				{
					FString Stats = FString::Printf
					(
						TEXT("snd=%02.1f recv=%02.1f"),
						GSecondsPerCycle*1000*Connection->Driver->SendCycles,
						GSecondsPerCycle*1000*Connection->Driver->RecvCycles
					);
					Connection->Actor->eventClientMessage( *Stats, NAME_None, 0 );
				}
			}
		}
		AccTickRate = 0;
		SkipCount = 0;
		ST_Traces = 0;
	}
	NumPV = 0;
	NumReps = 0;

	unguard;
}


//Burst forces actors to be replicated at full rate if not relevant
//Has to be used with extreme care as it increases check count
#define BURST_OWNED_COUNT 100
#define BURST_AR_COUNT 50

//Multithreaded relevancy builder (one of these days)
MS_ALIGN(16) struct FConnectionRelevancyList
{
	UXC_NetConnectionHack* Connection;	//O=0
	FConnectionRelevancyList* Next;	//O=4

	FActorPriority* PriorityArray;	//O=8
	FActorPriority** PriorityRefs;	//O=12

	AActor** ConsiderList;			//O=16
	INT ConsiderListSize;			//O=20
	INT OwnedConsiderListStart;		//O=24 (where our owned actors start in main consider list, not used in relevancy loop)
	INT OwnedConsiderListSize;		//O=28
	INT SpecialConsiderListSize;	//O=32

	FRotator Rotation;				//O=36
	FVector4 Location;				//O=48 (FVector4 has alignment=16)
	AActor* Viewer;					//O=64

	FConnectionRelevancyList( UNetConnection* InConn, FConnectionRelevancyList* InNext, INT InID)
		:	Connection( (UXC_NetConnectionHack*)InConn)
		,	Next( InNext)
	{};

	//This is a multithread nightmare, let's put it on main thread for now
	void GetViewerCoordinates()
	{
		// Get viewer coordinates.
		Location  = FVector4( Connection->Actor->Location, -1.f);
		Rotation  = Connection->Actor->ViewRotation;
		Viewer    = Connection->Actor;
		Connection->Actor->eventPlayerCalcView( Viewer, Location, Rotation );
		check(Viewer);
	}
	
	void PawnBandwidthHack( UActorChannel* Channel)
	{
		if ( !Channel->Recent.Num() )
			return;
		guard(PawnBandwidthHack);
		APawn* OldPawn = (APawn*)Channel->Recent.GetData();
		APawn* NewPawn = (APawn*)Channel->Actor;
		FLOAT Delta = Connection->Driver->Time - Channel->LastUpdateTime;
		FVector CalcLocation = OldPawn->Location + OldPawn->Velocity * Delta;
		if ( (CalcLocation - NewPawn->Location).SizeSquared() < 1.0f )
			OldPawn->Location = NewPawn->Location;
		unguard;
	}
	
	UClass* FindNetworkSuperClass( AActor* Actor) //Actor->bSuperClassRelevancy
	{
		UClass* Result = Actor->GetClass();
		while ( Result && (Result != AActor::StaticClass()) )
		{
			if ( (Result->ClassFlags & CLASS_Abstract) || (Connection->PackageMap->ObjectToIndex(Result) == INDEX_NONE) )
				Result = Result->GetSuperClass();
			else
				return Result;
		}
		return NULL;
	}

	
	void RelevancyLoop( FLOAT RealDelta, FLOAT MaxTickRate, INT ConnID)
	{
		guard(RelevancyLoop);
		INT ConsiderCount = 0;


		guard(SetupTraceStart);
		GetViewerCoordinates();
		if ( Connection->TickCount & 1 ) //Happens every 2 frames
		{
			FLOAT TMult = (Connection->TickCount & 2) ? 0.4f : 0.9f; //Pick 0.4 or 0.9 seconds (one frame 0.4, next 0.9 and so on)
			FVector4 Ahead = Location + Viewer->Velocity * TMult;
			if ( Viewer->Base ) //Add platform/lift velocity to the Ahead calc
				Ahead += Viewer->Base->Velocity * TMult;
			if ( ((UXC_Model*)Viewer->XLevel->Model)->FastRepLineCheck( &Ahead, &Location) )
				Location = Ahead;
		}
		unguard;

		TArray<AActor*>* ST = Connection->GetSentTemporaries();
		INT i;

		//Skip sent temporaries (set NetTag to negative)
		for ( i=0 ; i<ST->Num() ; i++ )
			(*ST)(i)->NetTag |= 0x80000000;

		FVector4 Dir( Rotation.Vector(), 0.f); //0.f is critical
		TMap<AActor*,UActorChannel*>* AC = Connection->GetActorChannels();
		TArray<UChannel*>* OC = Connection->GetOpenChannels();
		
		//Ez mode
		FLOAT OffsetFloat = (FLOAT)ConnID * 0.045;
		INT ChannelCount = OC->Num();
		INT ExpectedChannelCount = ChannelCount + Min(OwnedConsiderListSize, BURST_OWNED_COUNT);
		UBOOL bChannelsSaturated = ChannelCount > 800; //Change later!

		//Prioritize owned actors first
		guard(PrioritizeOwned);
		AActor** OwnedConsiderList = &ConsiderList[OwnedConsiderListStart];
		for ( i=0 ; i<OwnedConsiderListSize ; i++ )
		{
			AActor* Actor = OwnedConsiderList[i];
			if ( Actor->NetTag >= 0 )
			{
				UBOOL bPrioritize = 0;
				UActorChannel* Channel = AC->FindRef(Actor);

				if ( !Channel && (i < BURST_OWNED_COUNT) ) //Force prioritization when bursting
					bPrioritize = 1;
				else
				{
					if ( Actor->bAlwaysRelevant ) //Already passed general check time earlier
					{
						bPrioritize = !( Actor->CheckRecentChanges() && Channel && Channel->Recent.Num() && !Channel->Dirty.Num() && Actor->NoVariablesToReplicate( (AActor*) &Channel->Recent(0)) );
						if ( !bPrioritize ) //bAlwaysRelevant actor skips an update
							Channel->RelevantTime = Connection->Driver->Time;
					}
					else
					{
						FLOAT UF = Actor->UpdateFrequency( Viewer, Dir, Location);
						bPrioritize = (UF >= MaxTickRate) || (appRound( (*(FLOAT*)&Actor->NetTag + OffsetFloat)*UF) != appRound( (*(FLOAT*)&Actor->NetTag + OffsetFloat - RealDelta)* UF));
						//Owned actors shouln't spread updates, do not add to OffsetFloat
					}
				}

				if ( bPrioritize )
				{
					PriorityArray[ConsiderCount] = FActorPriority( EPri_Owned, Location, Dir, Connection, Channel, Actor);
					PriorityRefs[ConsiderCount] = PriorityArray + ConsiderCount;
					if ( bChannelsSaturated && (i >= BURST_OWNED_COUNT) && (PriorityArray[ConsiderCount].SuperRelevant == 1) && !Actor->bHidden && !Actor->Inventory && (Actor != Connection->Actor) )
						PriorityArray[ConsiderCount].SuperRelevant = 0; //We can afford to lose visible owned actors above burst capacity
					ConsiderCount++;
				}
			}
		}
		unguardf( (TEXT("[%i/%i]"), i, OwnedConsiderListSize));

		//Spectator: artificially reduce tickrate to half when updating non-owned stuff
		//Update owned stuff as usual, but update others at half the rate
		UBOOL bSkipMainPri = false; //Do not prioritize, simply fix NetTag
		if ( (Connection->UserFlags&32) || (Connection->Actor->PlayerReplicationInfo && Connection->Actor->PlayerReplicationInfo->bIsSpectator) )
		{
			if ( (Connection->TickCount & 1) == 0 ) //This will prevent the general loop
				bSkipMainPri = true;
			RealDelta *= 2.f;
			MaxTickRate *= 0.5f;
		}

		//If this is negative, then the channel list is expected to be FULL by end of this function
		INT DiscardCount = Max( 0, ExpectedChannelCount-923); //Use 100+(0 to FrameRate) channels as 'empty' buffer

		//Actors exist, have role
		guard(PrioritizeActors);
		i = bSkipMainPri ? ConsiderListSize : 0; //If skipping, set i=ConsiderListSize to avoid updating
		for ( ; i<SpecialConsiderListSize ; i++ ) //Process specials, super branchy code
		{
			AActor* Actor = ConsiderList[i];
			if ( Actor->NetTag > 0 )
			{
				FLOAT UF = Actor->NetUpdateFrequency;
				OffsetFloat += 0.023f;
				if ((UF >= MaxTickRate) || (appRound((*(FLOAT*)&Actor->NetTag + OffsetFloat)*UF) != appRound((*(FLOAT*)&Actor->NetTag + OffsetFloat - RealDelta)* UF)))
				{
					UBOOL SuperRelevancy = 1; //SuperRelevancy override
					UActorChannel* Channel = AC->FindRef(Actor);

					if (Actor->bRelevantIfOwnerIs) //Immediately close channel if owner isn't relevant
					{
						if (!Actor->Owner || !AC->FindRef(Actor->Owner))
						{
							if (Channel)
								Channel->Close();
							continue;
						}
					}

					if (Actor->bRelevantToTeam) //Can be combined with above condition
					{
						APlayerReplicationInfo* PRI = Connection->Actor->PlayerReplicationInfo;
						if (!PRI || PRI->bIsSpectator)
							continue;
						UByteProperty* TeamProperty = Cast<UByteProperty>(FindScriptVariable(Actor->GetClass(), TEXT("Team"), NULL));
						SuperRelevancy = -1;
						if (TeamProperty && (*(((BYTE*)Actor) + TeamProperty->Offset) == PRI->Team)) //Same team
						{
							SuperRelevancy = 1;
							//Treated as bAlwaysRelevant, can skip updates
							if (Actor->CheckRecentChanges() && Channel && Channel->Recent.Num() && !Channel->Dirty.Num() && Actor->NoVariablesToReplicate((AActor*)&Channel->Recent(0)))
							{
								Channel->RelevantTime = Connection->Driver->Time;
								continue;
							}
						}
						else if (!Channel) //No channel, don't even bother prioritizing
							continue;
					}

					if (Actor->IsOwnedBy(Connection->Actor))
						PriorityArray[ConsiderCount] = FActorPriority(EPri_Owned, Location, Dir, Connection, Channel, Actor);
					else
						PriorityArray[ConsiderCount] = FActorPriority(EPri_General, Location, Dir, Connection, Channel, Actor);
					PriorityArray[ConsiderCount].SuperRelevant = SuperRelevancy;
					PriorityRefs[ConsiderCount] = PriorityArray + ConsiderCount;
					ConsiderCount++;
				}
			}
		}
		
		INT Top = OwnedConsiderListStart;
		while ( i<ConsiderListSize )
		{
			for ( ; i<Top ; i++ )
			{
				AActor* Actor = ConsiderList[i];
				check(Actor);
				if ( Actor->NetTag > 0 )
				{
					UBOOL bPrioritize = 0;
					UActorChannel* Channel = AC->FindRef(Actor);
					if ( Actor->bAlwaysRelevant ) //Already passed general check time earlier
					{
						bPrioritize = !( Actor->CheckRecentChanges() && Channel && Channel->Recent.Num() && !Channel->Dirty.Num() && Actor->NoVariablesToReplicate( (AActor*) &Channel->Recent(0)) );
						if ( !bPrioritize ) //bAlwaysRelevant actor skips an update
							Channel->RelevantTime = Connection->Driver->Time;
					}
					else
					{
						FLOAT UF = Actor->UpdateFrequency( Viewer, Dir, Location);
						bPrioritize = (UF >= MaxTickRate) || (appRound( (*(FLOAT*)&Actor->NetTag + OffsetFloat)*UF) != appRound( (*(FLOAT*)&Actor->NetTag + OffsetFloat - RealDelta)* UF));
						OffsetFloat += 0.023f;
					}
	
					if ( bPrioritize )
					{
						PriorityArray[ConsiderCount] = FActorPriority( EPri_General, Location, Dir, Connection, Channel, Actor);
						PriorityRefs[ConsiderCount] = PriorityArray + ConsiderCount;
						if ( bChannelsSaturated && !Actor->bHidden && (Actor != Connection->Actor) )
							PriorityArray[ConsiderCount].SuperRelevant = 0; //We can afford to lose visible actors
						ConsiderCount++;
					}
				}
			}
			//Reached owner list, bypass it
			if ( Top == OwnedConsiderListStart )
				i = OwnedConsiderListStart + OwnedConsiderListSize;
			else
				i = ConsiderListSize;
			Top = ConsiderListSize;
		}
		unguardf( (TEXT("[%i/%i]"), i, ConsiderListSize));
		Connection->LastRepTime = Connection->Driver->Time;

		//SentTemporaries skipped, reset
		for ( i=0 ; i<ST->Num() ; i++ )
			(*ST)(i)->NetTag &= 0x7FFFFFFF;
		
		INT SuperRelevantCount = 0;

		guard(SortPriorities);

		//Sort by SuperRelevancy (2 blocks)
		{
			i=0;
			while ( i<ConsiderCount && (PriorityRefs[i]->SuperRelevant > 0) ) //Find first non-Super, use as buffer
				i++;
			SuperRelevantCount = i++;
			for ( ; i<ConsiderCount ; i++ ) //Arrange all other SuperRelevants into first block
				if ( PriorityRefs[i]->SuperRelevant > 0 )
					Exchange( PriorityRefs[i], PriorityRefs[SuperRelevantCount++] );
			for ( i=SuperRelevantCount ; i<ConsiderCount ; i++ ) //Arrange discarded actors last and leave them out
				if ( (PriorityRefs[i]->SuperRelevant < 0) && !PriorityRefs[i]->Channel )
					Exchange(PriorityRefs[i--], PriorityRefs[--ConsiderCount]);
		}

		//Sort by priority on each SuperRelevancy block
		INT HighBounds[2] = { SuperRelevantCount, ConsiderCount };
		INT LowBounds[2] = { 1, SuperRelevantCount+1 };
//		PriorityRefs = PriorityRefs_New;
		for ( INT j=0 ; j<2 ; j++ )
			for ( i=LowBounds[j] ; i<HighBounds[j] ; i++ ) //2 bound sets, one for each block
			{
				FActorPriority* CurPri = PriorityRefs[i]; //Buffered 'Current Priority being evaluated'
				INT k;
				for ( k=i ; (k>=LowBounds[j]) && (CurPri->Priority > PriorityRefs[k-1]->Priority) ; k-- ) //Find where 'Current' goes
					PriorityRefs[k] = PriorityRefs[k-1]; //Push elements up without swapping (faster!)
				PriorityRefs[k] = CurPri;
			}
		unguard;

		//Too many super relevants, discard less
		if ( ConsiderCount - DiscardCount < SuperRelevantCount )
			DiscardCount = ConsiderCount - SuperRelevantCount;

		UNetDriver* NetDriver = Connection->Driver;
		
		//Close 1 channel or discard every 0.25 second
		if ( DiscardCount > 0 )
		{
			guard(DiscardOne);
			INT FrameRate = Clamp( appRound(MaxTickRate) >> 2, 5, 50);
			if ( Connection->TickCount % FrameRate == 0 )
			{
				if ( PriorityRefs[ConsiderCount-1]->Channel )
					PriorityRefs[ConsiderCount-1]->Channel->Close();
				ConsiderCount--;
			}
			unguard;
		}

		INT Updated = 0;
		static FVector4 EndOffset; //Not thread safe, but 100% exception free
		EndOffset = FVector4(VRand(), 1.f);
		EndOffset.X *= 0.95;
		EndOffset.Y *= 0.95;
		EndOffset.Z *= 0.95;

		guard(Relevancy);
		for ( i=0 ; i<ConsiderCount ; i++ )
		{
			UActorChannel* Channel = PriorityRefs[i]->Channel;
			if ( !Channel && ChannelCount > 1022 ) //We don't need to perform a check if the list is saturated
				continue;

			AActor* Actor = PriorityRefs[i]->Actor;
			UBOOL IsRelevant = PriorityRefs[i]->SuperRelevant;
			if ( IsRelevant == -1 ) //Do not run relevancy checks
				IsRelevant = 0;
			else if ( !IsRelevant && (!Channel || (NetDriver->Time - Channel->RelevantTime > 0.3f)) )
				IsRelevant = ActorIsRelevant( Actor, Connection->Actor, Viewer, &Location, EndOffset);

			
			if( IsRelevant || (Channel && NetDriver->Time-Channel->RelevantTime < NetDriver->RelevantTimeout) )
			{
				Actor->XLevel->NumPV++;
				
				if( !Channel )
				{
					UClass* BestClass = Actor->GetClass();
					if ( (Connection->PackageMap->ObjectToIndex(BestClass) != INDEX_NONE) ||
						(Actor->bSuperClassRelevancy && ((BestClass=FindNetworkSuperClass(Actor)) != NULL)) )
					{
						Channel = (UActorChannel*)Connection->CreateChannel( CHTYPE_Actor, 1, INDEX_NONE);
						if ( Channel )
						{
							if ( BestClass != Actor->GetClass() ) //SuperHack
							{
								UClass** ClassAddr = ((UClass**)Actor) + 9; //Offset 36
								Exchange( BestClass, ClassAddr[0]);
								Channel->SetChannelActor( Actor );
								Exchange( BestClass, ClassAddr[0]);
							}
							else
								Channel->SetChannelActor( Actor );
							ChannelCount++;
						}
					}
/*					if ( Connection->PackageMap->ObjectToIndex(BestClass) != INDEX_NONE )
					{
						// Create a new channel for this actor.
						Channel = (UActorChannel*)Connection->CreateChannel( CHTYPE_Actor, 1, INDEX_NONE);
						if( Channel )
						{
							Channel->SetChannelActor( Actor );
							ChannelCount++;
						}
					}*/
				}
				if ( Channel )
				{
					if ( !Connection->IsNetReady(0) )
						break;
					if( IsRelevant )
						Channel->RelevantTime = NetDriver->Time + 0.1 + appFrand();
					if( Channel->IsNetReady(0) )
					{
						//Hack for simulated pawns, reduces bandwidth
						if ( Actor->bSimulatedPawn && (Actor->Physics != PHYS_None) )
							PawnBandwidthHack( Channel);
						if ( Channel->ActorClass != Actor->GetClass() ) //Custom rep rules!
						{
							UClass* BestClass = Channel->ActorClass;
							UClass** ClassAddr = ((UClass**)Actor) + 9; //Offset 36
							Exchange( BestClass, ClassAddr[0]);
							Channel->ReplicateActor();
							Exchange( BestClass, ClassAddr[0]);
						}
						else
							Channel->ReplicateActor();
						Updated++;
					}
					if ( !Connection->IsNetReady(0) )
						break;
				}
			}
			else if ( Channel && (Actor != Connection->Actor)) //Never close the local player's channel
				Channel->Close();
				
		}

		//Saturation flag
		if ( i < ConsiderCount )
			Connection->UserFlags |= 32;
		else
			Connection->UserFlags &= ~32;
		
		//Make sure channels don't go past RelevantTimeout if they don't have to during saturation
		for ( ; i<ConsiderCount ; i++ )
		{
			UActorChannel* Channel = PriorityRefs[i]->Channel;
			if ( Channel && (NetDriver->Time-Channel->RelevantTime < NetDriver->RelevantTimeout - 1.1f) )
			{
				AActor* Actor = PriorityRefs[i]->Actor;
				UBOOL IsRelevant = PriorityRefs[i]->SuperRelevant;
				if ( IsRelevant == -1 ) //Do not run relevancy checks
					IsRelevant = 0;
				else if ( !IsRelevant )
					IsRelevant = ActorIsRelevant( Actor, Connection->Actor, Viewer, &Location, EndOffset);
				//Reset the relevancy timers if the channels are saturated
				if ( IsRelevant )
					Channel->RelevantTime = NetDriver->Time + (1.1f - NetDriver->RelevantTimeout);
			}
		}
		
		unguardf( (TEXT("[%i/%i] %s"), i, ConsiderCount, PriorityRefs[i]->Actor->GetName() ));
	
		
		unguard;
	}


} GCC_ALIGN(16);


INT UXC_Level::ServerTickClients( FLOAT DeltaSeconds )
{
	guard(UXC_Level::ServerTickClients);

	DeltaSeconds *= _Reciprocal( GetLevelInfo()->TimeDilation);

	INT i;
	INT Updated=0;
	FMemMark Mark(GMem);
	FConnectionRelevancyList* ConnList = NULL;

	guard( MakeConnectionLists);
	for( i=NetDriver->ClientConnections.Num()-1; i>=0; i-- ) //Reverse order, so chained list isn't flipped
	{
		UNetConnection* Conn = NetDriver->ClientConnections(i);
		check( Conn);
		check( Conn->State==USOCK_Pending || Conn->State==USOCK_Open || Conn->State==USOCK_Closed);

		if ( !Conn->Actor || !Conn->IsNetReady(0) || Conn->State!=USOCK_Open )
			continue;
		if ( Conn->Driver->Time - Conn->LastReceiveTime > 1.5f )
			continue;
		ConnList = new(GMem,1,16) FConnectionRelevancyList( Conn, ConnList, Updated);
		Updated++;
	}
	//No clients, abort here
	if ( !Updated )
		return 0;
	unguard;

	AActor** ConsiderList;
	INT ConsiderListSize = 0;
	INT NO_ConsiderListSize = 0;
	INT NO_OwnedConsiderListSize = 0;
	
	guard( MakeConsiderList);
	INT ActorListSize = Actors.Num() + 2 + 1 - iFirstNetRelevantActor; //That +1 is a buffer!!!
	INT OwnedListPos = ActorListSize;
	ConsiderList = new(GMem, ActorListSize+1) AActor*;
	// Add LevelInfo
	if( Actors(0) && (Actors(0)->RemoteRole!=ROLE_None) )
	{
		FLOAT OldTime = UpdateNetTag( Actors(0), DeltaSeconds);
		if ( appRound( OldTime*Actors(0)->NetUpdateFrequency) != appRound( *((FLOAT*)&Actors(0)->NetTag)*Actors(0)->NetUpdateFrequency) )
		{
			Actors(0)->bAlwaysRelevant = true; //Important to enforce this
			*(FLOAT*) &Actors(0)->NetTag += appFrand() * 0.05; //Randomizer
			ConsiderList[ConsiderListSize++] = Actors(0);
		}
	}
	//Non-owned actors from (0 to max), owned actors from (max to 0)
	FLOAT ArOffset = 0.f; //bAlwaysRelevant actors spreader
	INT SpecialConsiders = 0;
	for( i=iFirstNetRelevantActor; i<Actors.Num(); i++ ) //Consider all actors for now
	{
		AActor* Actor = Actors(i);
		if ( Actor && Actor->RemoteRole != ROLE_None )
		{
			//NetTag rules: always positive, between 0 and 99 (up to 120)
			FLOAT OldTime = UpdateNetTag( Actor, DeltaSeconds);
			UBOOL bPass = 0;
			
			//Special actors go first in the list, push any other out to last!
			if ( Actor->bRelevantIfOwnerIs || Actor->bRelevantToTeam )
			{
				ConsiderList[ConsiderListSize++] = ConsiderList[SpecialConsiders];
				ConsiderList[SpecialConsiders++] = Actor;
			}
			else if ( Actor->bAlwaysRelevant )
			{
				//bAlwaysRelevant actors are checked at the same time
				if ( appRound( (OldTime+ArOffset)*Actor->NetUpdateFrequency) != appRound( (*((FLOAT*)&Actor->NetTag)+ArOffset)*Actor->NetUpdateFrequency) )
				{
					*(FLOAT*) &Actor->NetTag += appFrand() * 0.05; //Randomizer (is it necessary?)
					ArOffset += 0.023f;
					bPass = 1;
				}
			}
			else if ( !(Actor->bNetOptional && (Actor->GetClass()->GetDefaultActor()->LifeSpan - 0.15f >= Actor->LifeSpan)) )
				bPass = 1;

			if ( bPass )
			{
				if ( OwnedByAPlayer(Actor) )
					ConsiderList[--OwnedListPos] = Actor;
				else
					ConsiderList[ConsiderListSize++] = Actor;
			}
		}
	}

	NO_ConsiderListSize = ConsiderListSize;
	//Sort owned list by ConnectionID and move to end of ConsiderList incrementally
	for ( FConnectionRelevancyList* ConnLink=ConnList ; ConnLink ; ConnLink=ConnLink->Next )
	{
		ConnLink->OwnedConsiderListStart	= ConsiderListSize;
		ConnLink->OwnedConsiderListSize		= 0;
		ConnLink->SpecialConsiderListSize	= SpecialConsiders;
		APlayerPawn* NetOwner = ConnLink->Connection->Actor;
		//Grab all owned actors
		for ( i=OwnedListPos ; i<ActorListSize ; i++ )
			if ( ConsiderList[i]->IsOwnedBy(NetOwner) )
			{
				ConsiderList[ConsiderListSize++] = ConsiderList[i];
				ConsiderList[i] = ConsiderList[OwnedListPos++];
				ConnLink->OwnedConsiderListSize++;
			}
	}

	//Move any leftovers not belonging to connected players (filter them out?, useful for bot games)
	//Ideally, it's zero moves
	while ( OwnedListPos < ActorListSize )
	{
		ConsiderList[ConsiderListSize++] = ConsiderList[OwnedListPos++];
		NO_OwnedConsiderListSize++;
	}
	unguard;


	guard( ArrayProcess);

	FLOAT MaxTickRate = Engine->GetMaxTickRate(); //This should become an argument
	if ( MaxTickRate == 0 && (DeltaSeconds != 0.f) )
		MaxTickRate = _Reciprocal( DeltaSeconds);
	//Allocate general chunks, they will be reused by each connection
	FActorPriority* GPriorityArray = new(GMem,ConsiderListSize) FActorPriority;
	FActorPriority** GPriorityRefs = new(GMem,ConsiderListSize) FActorPriority*;

	i = 1;
	for ( FConnectionRelevancyList* ConnLink=ConnList ; ConnLink ; ConnLink=ConnLink->Next )
	{
		ConnLink->PriorityArray		= GPriorityArray;
		ConnLink->PriorityRefs		= GPriorityRefs;
		ConnLink->ConsiderList		= ConsiderList;
		ConnLink->ConsiderListSize	= ConsiderListSize;
		ConnLink->Connection->TickCount++;
		NetTag++;
		ConnLink->RelevancyLoop( DeltaSeconds, MaxTickRate, i-1); //Internal processing starts with 0, not 1
		i++;
	}
	unguardf( (TEXT("[%i/%i] NOG=%i, NOW=%i, NOO=%i"), i, Updated, NO_ConsiderListSize, ConsiderListSize-NO_ConsiderListSize, NO_OwnedConsiderListSize));

	Mark.Pop();
	return Updated;
	unguard;
}

//This doesn't appear to be working, dammit
//Prevents heavy log spamming
//TODO: Figure out a way to allow a client to rebroadcast a game?
EAcceptConnection UXC_Level::NotifyAcceptingConnection()
{
	check(NetDriver);
	static FTime LastNAC = 0.f; //Don't initialize a static variable using appSeconds(), just in case
	FTime CurTime = TimeSeconds;
	if( NetDriver->ServerConnection )
	{
		// We are a client and we don't welcome incoming connections.
		// Don't fill a client's log
		if ( CurTime - LastNAC > 1.f )
		{
			LastNAC = CurTime;
			debugf( NAME_DevNet, TEXT("NotifyAcceptingConnection: Client refused") );
		}
		return ACCEPTC_Reject;
	}
	else if( GetLevelInfo()->NextURL!=TEXT("") )
	{
		// Server is switching levels.
		// 100% useless, populated v436 servers suffer a lot from this
//		debugf( NAME_DevNet, TEXT("NotifyAcceptingConnection: Server %s refused"), GetName() );
		return ACCEPTC_Ignore;
	}
	else
	{
		// Server is up and running.
		// Log everything during the first second of the level
		// Otherwise add a 0.25s interval to avoid DDoS's from filling log
		if ( (NetDriver->Time.GetFloat() < 1.f) || (CurTime - LastNAC > 0.25f) )
		{
			LastNAC = CurTime;
			debugf( NAME_DevNet, TEXT("NotifyAcceptingConnection: Server %s accept"), GetName() );
		}
		return ACCEPTC_Accept;
	}
}

void UXC_Level::NotifyAcceptedConnection( class UNetConnection* Connection )
{
	Super::NotifyAcceptedConnection( Connection);
}

//Detaches parameter from login
static UBOOL RemoveParam( FString& Str, FString& Parsed, const TCHAR* ParamName)
{
	guard(RemoveParam);
	INT PLen = appStrlen( ParamName);
	if ( !PLen || (Str.Len() == 0) )
		return 0;
	INT Location = 0;
	INT STLen = Str.Len();
	while ( Location < STLen)
	{
		while ( (*Str)[Location] == '?' )
			Location++;
		if ( Str.Mid(Location, PLen) == ParamName )
		{
			//Parameter validated, extract!
			Parsed = Str.Mid( Location);
			Str = Str.Left( Location);
			INT NewLocation = Parsed.InStr( TEXT("?") );
			if ( NewLocation > 0 )
			{
				Str += Parsed.Mid( NewLocation+1); //Get rid of a ? char
				Parsed = Parsed.Left( NewLocation);
			}
			return 1;
		}
		while ( (Location < STLen) && ((*Str)[Location] != '?') )
			Location++;
	}
	return 0;
	unguard;
}

static TCHAR HelloText[]	= { 'H', 'E', 'L', 'L', 'O', 0 };
static TCHAR JoinText[]		= { 'J', 'O', 'I', 'N', 0 };
static TCHAR LoginText[]	= { 'L', 'O', 'G', 'I', 'N', 0 };
static TCHAR UserFlagText[]	= { 'U', 'S', 'E', 'R', 'F', 'L', 'A', 'G', 0 }; //Do not userflag pre-join
static TCHAR HaveText[]		= { 'H', 'A', 'V', 'E', 0 };
static TCHAR SkipText[]		= { 'S', 'K', 'I', 'P', 0 };
static TCHAR NetspeedText[]	= { 'N', 'E', 'T', 'S', 'P', 'E', 'E', 'D', 0 };

static TCHAR* ValidClientCommands[] = { UserFlagText, NetspeedText };
static TCHAR* ValidJoinCommands[] = { HelloText, JoinText, LoginText, HaveText, SkipText, NetspeedText};

void UXC_Level::NotifyReceivedText(UNetConnection* Connection, const TCHAR* Text)
{
	guard(UXC_Level::NotifyReceivedText);
	if ( NetDriver->ServerConnection )
	{
	}
	else
	{
		if (Connection->State == USOCK_Closed)
			return;

		const TCHAR* Str = Text;
		//Only allow post player join commands after join is complete
		if ( Connection->Actor )
		{
			UBOOL bFail = 1;
			for ( INT i=0 ; i<ARRAY_COUNT(ValidClientCommands) ; i++ )
				if ( ParseCommand( &Str, ValidClientCommands[i]) )
				{
					bFail = 0;
					break;
				}
			if ( bFail )
				return;
		}
		//Only allow player join commands when needed
		else
		{
			UBOOL bFail = 1;
			for (INT i = 0; i<ARRAY_COUNT(ValidJoinCommands); i++)
				if (ParseCommand(&Str, ValidJoinCommands[i]))
				{
					bFail = 0;
					break;
				}
			if (bFail)
				return;
		}

		Str = Text;

		if (ParseCommand(&Str, TEXT("LOGIN")))
		{
			if ( Connection->RequestURL.Len() )
				return;
		}
		else if ( ParseCommand(&Str, TEXT("JOIN")) )
		{
			// See that RequestURL is appropiate before doing any extra processing
			FString LoginParams = Connection->RequestURL;
			FString Name;
			FString Class;

			FString Error;
			if ( LoginParams == TEXT("") ) //Did player even LOGIN?
				Error = TEXT("[XCGE] LOGIN command invalid or not received");
			else if ( !RemoveParam( LoginParams, Name, TEXT("Name=")) ) //Player has a name?
				Error = TEXT("[XCGE] No player name in login request");
			else if ( !RemoveParam( LoginParams, Class, TEXT("Class=")) ) //Player has a class?
				Error = TEXT("[XCGE] No player class in login request");

			// Finish computing the package map.
			if ( !Error.Len() ) //No error
				Connection->PackageMap->Compute();

			debugf(NAME_DevNet, TEXT("Join request: %s"), *Connection->RequestURL);
			// Spawn the player-actor for this network player.
			if ( Error.Len() || !SpawnPlayActor(Connection, ROLE_AutonomousProxy, FURL(NULL, *Connection->RequestURL, TRAVEL_Absolute), Error) )
			{
				// Failed to connect.
				debugf(NAME_DevNet, TEXT("Join failure: %s"), *Error);
				Connection->Logf(TEXT("FAILURE %s"), *Error);
				Connection->FlushNet();
				Connection->State = USOCK_Closed;
			}
			else
			{
				// Successfully in game.
				debugf(NAME_DevNet, TEXT("Join succeeded: %s"), *Connection->Actor->PlayerReplicationInfo->PlayerName);
			}
			return;
		}
	}

	Super::NotifyReceivedText( Connection, Text);
	unguard;
}

//====================================================
//====================================================
// Brush tracker fixer
//====================================================
//====================================================

#ifndef DISABLE_ADDONS

UBOOL FXC_BrushTrackerFixer::Init()
{
	guard( FXC_BrushTrackerFixer::Init);
	check( XLevel);

	if ( XLevel->BrushTracker || !XLevel->Actors(0) )
		return false;

	debugf( NAME_XC_Engine, TEXT("Checking for Moving Brush Tracker insertion..."));
	for( INT i=0; i<XLevel->Actors.Num(); i++ )
	{
		guard(MoverCheck);
		if( StaticMover( XLevel->Actors(i)) ) //Static mover
		{
			if ( !XLevel->BrushTracker )
				XLevel->BrushTracker = GNewBrushTracker( XLevel );
			XLevel->BrushTracker->Flush( XLevel->Actors(i) );
			XLevel->BrushTracker->Update( XLevel->Actors(i) );
			StaticMovers.AddItem( (AMover*) XLevel->Actors(i) );
		}
		unguard;
	}
	if ( !StaticMovers.Num() )
		return false;
	StaticMovers.Shrink();
	return true;

	unguard;
}

void FXC_BrushTrackerFixer::Exit()
{
	if ( StaticMovers.Num() > 0 )
		StaticMovers.Empty();
}

UBOOL FXC_BrushTrackerFixer::IsTyped( const TCHAR* Type)
{
	return appStricmp( Type, TEXT("BrushTrackerFixer")) == 0;
}

INT FXC_BrushTrackerFixer::Tick( FLOAT DeltaSeconds)
{
	guard( FXC_BrushTrackerFixer::Tick)
	if ( !XLevel || StaticMovers.Num() <= 0 || !XLevel->BrushTracker )
		return 0; //Should deinitialize
	if ( XLevel->Actors.Num() <= 0 || !XLevel->Actors(0) || !XLevel->GetLevelInfo()->bBegunPlay )
		return 1; //Should not deinitialize
	for( INT i=0; i<StaticMovers.Num(); i++ )
	{
		if ( !StaticMovers(i) || StaticMovers(i)->bDeleteMe || StaticMovers(i)->bHidden )
			StaticMovers.Remove(i);
		else
			XLevel->BrushTracker->Update( StaticMovers(i) );
	}
	//TestLevel->BrushTracker->Flush( TestLevel->Actors(i) );

	return 1;
	unguard;
}

#endif


